(in-package "GWB")

(defparameter *simple-constants*
  `((+ 2 ,#'+)
    (- 2 ,#'-)
    (* 2 ,#'*)
    (true 0 true)
    (false 0 false)
    (= 2 ,#'(lambda (x y) (if (= x y) 'true 'false)))
    (print 1 ,#'(lambda (x) (format t "~&==> ~S~%" x) x)))
  "List of the constants for ground constants and strict functions
of the source language. Entries are of the form (NAME ARITY DEFINITION)")

(defparameter *constants*
  `((compose ,#'make-composed-strategy)
    (I ,#'make-I-strategy)
    (K ,#'make-K-strategy)
    (S ,#'make-S-strategy)
    (Y ,#'make-Y-strategy)
    (if ,#'make-conditional-strategy)
    (omega ,#'make-bottom-strategy)
    (top ,#'make-top-strategy)
    (call/cc ,#'make-call/cc-strategy)
    (set ,#'make-setter-strategy)
    (deref ,#'make-read-strategy)
    (skip ,#'make-skip-strategy)
    (seq ,#'make-seq-strategy)
    (pair ,#'make-pair-strategy)
    (p0 ,#'make-p0-strategy)
    (p1 ,#'make-p1-strategy))
  "List of constants with special-case code.
Entries are of the form (NAME CONSTRUCTOR)")

(defparameter *sugar*
  '((trace (t nil) t)
    (remember (t nil) t)
    (primitive-new (nil) t)
    (curry (t) t))
  "The list of syntactic sugar operators")

(defparameter *macros*
  `((begin
     ,#'(lambda (form)
          (let ((args (cdr form)))
          (unless (consp args)
            (error 'gwb-syntax-error :form form))
          (if (null (cdr args))
              (car args)
            `(seq ,(car args) (begin ,@(cdr args)))))))
    (new
     ,#'(lambda (form)
          (let ((name (cadr form))
                (initial-value (caddr form))
                (body (cadddr form)))
            `((lambda ,name ,body) (primitive-new ,initial-value)))))))

(defun toCL (form)
  "Translate from lambda to SKIY"
  (cond
    ((atom form) form)
    ((assoc (car form) *macros*)
     (let ((defn (assoc (car form) *macros*)))
       (toCL (funcall (cadr defn) form))))
    ((eql (car form) 'lambda)
     (multiple-value-bind (var body)
         (parse-lambda form)
       (toCL-lambda var (toCL body))))
    ((eql (car form) 'constant)
     (list 'constant (toCL (cadr form))))
    ((assoc (car form) *sugar*)
     (let* ((defn (assoc (car form) *sugar*))
            (CL
             (cons (car form)
                   (mapcar
                    #'(lambda (x y) (if y (toCL x) x))
                    (cdr form)
                    (cadr defn)))))
       CL))
    ((null (cdr form))
     (toCL (car form)))
    ((null (cddr form))
     (list (toCL (car form)) (toCL (cadr form))))
    (t
     (toCL (cons
            (list (car form) (cadr form))
            (cddr form))))))

(defun parse-lambda (form)
  (labels
      ((parse-lambda* (var varlist body)
         (let ((new-body (if varlist `(lambda ,varlist ,body) body)))
           (cond
             ((symbolp var) (values var new-body))
             (t (error 'gwb-syntax-error))))))
    (unless (= 3 (length form))
      (error 'gwb-syntax-error :form form))
    (let ((varlist (cadr form)) (body (caddr form)))
      (cond
        ((symbolp varlist) (values varlist body))
        ((listp varlist)
         (parse-lambda* (car varlist) (cdr varlist) body))
        (t (error 'gwb-syntax-error))))))

(defun toCL-lambda (var body)
  "Helper for toCL"
  (declare (symbol var))
  (cond
    ((eql var body) 'I)
    ((atom body) `(constant ,body))
    ((assoc (car body) *macros*)
     (let ((defn (assoc (car body) *macros*)))
       (toCL-lambda var (funcall (cadr defn) body))))
    ((eql (car body) 'constant)
     (toCL-lambda var `(K ,(cadr body))))
    ((assoc (car body) *sugar*)
     (let* ((defn (assoc (car body) *sugar*))
            (CL
             (cons (car body)
                   (mapcar
                    #'(lambda (x y)
                        (when y
                          (unless (and (not (eql var x)) (atom x))
                            (error 'gwb-syntax-error)))
                        x)
                    (cdr body)
                    (cadr defn)))))
       (if (caddr defn)
           `(constant ,CL)
         CL)))
    (t
     `((S ,(toCL-lambda var (car body)))
       ,(toCL-lambda var (cadr body))))))

(defun compile-CL (form)
  (let ((s (compile-CL* form)))
    s))

(defun compile-constant (name)
  (declare (symbol name))
  (let ((defn (assoc name *simple-constants*)))
    (if defn
        (let* ((value (caddr defn))
               (arity (cadr defn)))
          (ecase arity
            (0 (make-ground-strategy
                :value value))
            (1 (make-ground->ground-strategy
                :value value))
            (2 (make-ground3-strategy
                :value value))))
      (let ((defn (assoc name *constants*)))
        (if defn
            (let ((constructor (cadr defn)))
              (funcall constructor))
          nil)))))
                       
(defvar *remembered* '()
  "A-list of remembered strategies")

(defvar *environment* '()
  "A-list of defined terms")

(defun lookup (f)
  "Lookup a defined term"
  (declare (symbol f))
  (the (or null strategy) (cdr (assoc f *environment*))))

(defun compile-CL* (form)
  "Compile a SKIY term into a strategy"
  (cond
    ((integerp form)
     (make-ground-strategy :value form))
    ((symbolp form)
     (or (compile-constant form)
         (let ((s (lookup form)))
           (if s
               (fresh-strategy s)
             (error 'gwb-cell-error :kind "constant" :name form)))))
    ((consp form)
     (cond
       ((eql 'constant (car form))
        (constant-strategy (compile-CL (cadr form))))
       ((eql 'trace (car form))
        (trace-strategy
         (compile-CL (cadr form))
         (caddr form)))
       ((eql 'remember (car form))
        (let ((strat (compile-CL (cadr form))))
          (setf *remembered* (acons (caddr form) strat *remembered*))
          strat))
       ((eql 'primitive-new (car form))
        (make-new-strategy :value (make-value-token (cadr form))))
       ((eql 'curry (car form))
        (curry-strategy (compile-CL (cadr form))))
       (t
        (apply-strategy
         (compile-CL (car form)) (compile-CL (cadr form))))))
    (t (error 'gwb-syntax-error :form form))))

(defun compile-calculus (form)
  "Compile a form into a strategy"
  (compile-CL (toCL form)))

(defun curry-strategy (s)
  (declare (type strategy s))
  (make-curried-strategy :value s))

(defun decompile-strategy (s)
  (declare (type strategy s))
  (typecase s
    (bottom-strategy 'omega)
    (ground-strategy (ground-strategy-value s))
    (ground->ground-strategy 'f1)
    (ground3-strategy 'f2)
    (I-strategy 'I)
    (K-strategy 'K)
    (S-strategy 'S)
    (Y-strategy 'Y)
    (pair-strategy 'pair)
    (proj-strategy
     (intern (format nil "P~A" (proj-strategy-component s))))
    (conditional-strategy 'if)
    (un-constant-strategy
     (let ((s1 (un-constant-strategy-value s)))
       (if (and (composed-strategy-p s1)
                (constant-strategy-p (composed-strategy-s1 s1)))
           (list
            (decompile-strategy (composed-strategy-s2 s1))
            (decompile-strategy (constant-strategy-value
                                 (composed-strategy-s1 s1))))
         (list 'un-constant-strategy (decompile-strategy s1)))))
    (constant-strategy
     (list 'constant (decompile-strategy (constant-strategy-value s))))
    (curried-strategy
     (list 'curry (decompile-strategy (curried-strategy-value s))))
    (composed-strategy
     (list
      'compose
      (decompile-strategy (composed-strategy-s2 s))
      (decompile-strategy (composed-strategy-s1 s))))
    (t '?)))
    
    
