(in-package "GWB")

(declaim (inline comp-car))
(defun comp-car (comp)
  "Return the first projection in a component"
  (declare (type comp comp))
  (and (> (length comp) 0)
       (aref comp 0)))

;;; various functions on moves
(declaim (inline where))
(defun where (m)
  "Return the component of a move"
  (declare (type move m))
  (if (zerop (length (move-comp m)))
      (error 'gwb-root-move-error :move m))
  (aref (move-comp m) 0))

(defmacro wherep (m &rest c)
  "Test for a move being in a given component"
  (let ((comp-name (gensym "COMP")))
    (if (zerop (length c))
        `(progn ,m t)
      `(let ((,comp-name (move-comp ,m)))
        (and (>= (length ,comp-name) ,(length c))
        ,@(case (length c)
           (1 `((eql (aref ,comp-name 0) ,(car c))))
           (2 `((eql (aref ,comp-name 0) ,(car c))
                (eql (aref ,comp-name 1) ,(cadr c))))
           (3 `((eql (aref ,comp-name 0) ,(car c))
                (eql (aref ,comp-name 1) ,(cadr c))
                (eql (aref ,comp-name 2) ,(caddr c))))
           (t
            (let ((comp (coerce c 'comp)))
              `(wherep* ,m ,comp)))))))))

(defun wherep* (m c)
  "Function version of WHEREP"
  (declare (type move m) (type comp c))
  #.(declare-fast)
  (do ((comp (move-comp m)) (n 0 (+ n 1)))
      ((or (>= n (length comp)) (>= n (length c))
           (not (eql (aref comp n) (aref c n))))
       (>= n (length c)))))

(declaim (inline where-comp))
(defun where-comp (m &optional inverse)
  "Return the first projection of the component of a move"
  (declare (type move m))
  (if (zerop (length (move-comp m)))
      (error 'gwb-root-move-error :move m))
  (if (eql (aref (move-comp m) 0) 0)
      (if inverse #*1 #*0)
    (if inverse #*0 #*1)))

(defun comp-set-prefix (comp prefix)
  "Mutate COMP in-place so that it starts with PREFIX"
  (declare (type comp comp prefix))
  #.(declare-fast)
  (do ((n 0 (+ n 1)))
      ((>= n (length prefix)) comp)
    (declare (type type-int n))
    (setf (aref comp n) (aref prefix n))))

(defun comp-append (c1 c2 &optional (n 0))
  "Concatenate C1 and C2 omitting the first N projections of C2"
  (declare (type comp c1 c2) (type type-int n))
  #.(declare-fast)
  (cond 
   ((and (zerop n) (zerop (length c1))) c2)
   ((and (zerop n) (zerop (length c2))) c1)
   (t
    (do* ((l1 (length c1)) (l2 (length c2))
          (len (+ l1 (- l2 n)))
          (r (make-array len :element-type 'bit))
          (m 0 (+ m 1)))
        ((>= m len) r)
      (declare (type type-int m l1 l2 len))
      (setf (aref r m) (if (< m l1) (aref c1 m) (aref c2 (+ (- m l1) n))))))))

(declaim (inline changecomp))
(defun changecomp (m c1 c2 just)
  "Return a copy of a move in a different component and with a different \
justification"
  (declare (type move m) (type comp c1 c2) (type just just))
  (make-move
   (comp-append c2 (move-comp m) (length c1))
   (move-token m)
   just))

(defun comp-< (c1 c2)
  "Compare components lexicographically"
  (declare (type comp c1 c2))
  (cond
    ((zerop (length c2)) nil)
    ((zerop (length c1)) t)
    (t
     (do ((n 0 (+ n 1)))
         ((or (>= n (length c2))
              (>= n (length c1))
              (/= (aref c1 n) (aref c2 n)))
          (cond
            ((>= n (length c2)) nil)
            ((>= n (length c1)) t)
            (t (< (aref c1 n) (aref c2 n)))))))))
       

(defun comp-cons (head tail)
  "Construct a component from a projection and a component"
  (declare (type comp tail) (type bit head))
  #.(declare-fast)
  (let* ((l (length tail))
         (new-comp (make-array (+ 1 l) :element-type 'bit)))
    (setf (aref new-comp 0) head)
    (if (zerop l)
        new-comp
    (do ((n 0 (+ n 1)))
        ((>= n l) new-comp)
      (declare (type type-int n))
      (setf (aref new-comp (+ n 1)) (aref tail n))))))

(defconstant *question-tokens* '(?)
  "The list of tokens that are questions")

(defconstant *initial-question* '?
  "The initial question token")

(declaim (inline initial-move-p))
(defun initial-move-p (m)
  (null (move-just m)))

(defmacro questionp (move)
  "Checks for a move being a question"
  (let ((move-name (gensym "MOVE"))
        (token-name (gensym "TOKEN")))
    `(let* ((,move-name ,move)
            (,token-name (move-token ,move-name)))
       (declare (type move ,move-name))
       (or ,@(mapcar
              #'(lambda (tok) `(eql ,token-name ',tok))
              *question-tokens*)))))

(defun justified-by (move comp)
  "Finds the hereditary justifier of a move in a given component"
  (declare (type (or null move) move) (type comp comp))
  (if (null move) nil
    (if (wherep* move comp)
        move
      (justified-by (move-just move) comp))))

(defun find-move (move history)
  "Find a move in a history"
  (declare (type move move) (type hist history))
  #.(declare-fast)
  (do ((h history (cdr h)))
      ((or (null h) (eql move (car h)))
       h)))

(defun move-position (move history)
  (declare (type move move) (type hist history))
  #.(declare-fast)
  (do ((h history (cdr h)) (n 0 (+ n 1)))
      ((or (null h) (eq move (car h)))
       (if (null h)
           nil
         n))
    (declare (type hist-int n))))

(defvar *genname-counter* 2)
(declaim (type (integer 2 *) *genname-counter*))

(defun genname (&optional (prefix "anon") (suffix ""))
  (declare (type simple-string prefix suffix))
  (prog1
      (the string (format nil "~A~A~A" prefix *genname-counter* suffix))
    (incf *genname-counter*)))

(defun gennum ()
  (prog1
      *genname-counter*
    (incf *genname-counter*)))

