(in-package "GWB")

(defvar *gwb-handle-errors* t)

(defvar *gwb-prompt* "GWB> ")

(defun newdef (name s)
  (declare (type strategy s))
  (let ((olddef (assoc name *environment*)))
    (if olddef
        (setf (cdr olddef) s)
      (setf *environment* (acons name s *environment*)))))

(defun lookup-remembered (n)
    (let ((entry (assoc n *remembered*)))
      (when (null entry)
        (error 'gwb-cell-error :kind "remembered run" :name n))
      (cdr entry)))

(defun compile-term (term)
  (compile-calculus term))
  
(defun define-term (n l)
  (when (not (symbolp n))
    (error 'gwb-syntax-error :form l))
  (let ((strat (compile-term l)))
    (newdef n strat)
    (format t "~&~A~%" n)))

(defun run-term (l stream)
  (declare (type stream stream))
  (let ((strat (compile-term l)))
    (let ((nm (run-strategy strat)))
      (declare (fixnum nm))
      (format stream "~&~A~%" (strategy-returned-value strat))
      (format stream "~&~D moves played.~%" nm))))

(defun give-help (stream)
  (declare (special *directives*))    ; mutual recursion
  (declare (type stream stream))
  (dolist (entry *directives*)
    (if (null (cadr entry))
        (format stream "~&~A~30T~A~%"
                (car entry)
                (caddr entry))
      (format stream "~&(~A~{ ~A~})~@[~30T~A~]~%"
              (car entry)
              (cadr entry)
              (caddr entry)))))

(defparameter *directives*
  `((quit nil "Quit the GWB"
     ,#'(lambda () (throw 'gwb-quit nil)))
    (help nil "Display help" ,#'(lambda () (give-help *standard-output*)))
    (clear nil "Clear all definitions"
     ,#'(lambda () (setf *environment* '())))
    (list nil "List all definitions"
     ,#'(lambda ()
          (dolist (entry *environment*)
            (format t "~&~A%" (car entry)))))
    (forget nil "Forget all remembered runs"
     ,#'(lambda () (setf *remembered* '())))
    (define (name term) "Define a name to a term" ,#'define-term)
    (run (term) "Simulate the run of a term"
     ,#'(lambda (term) (run-term term *standard-output*)))
    (load (filename) "Load a file"
     ,#'(lambda (fn)
          (block load
            (let ((*gwb-prompt* nil))
              (with-open-file (infile fn :direction :input)
                (loop
                 (when (not (repl-1 infile))
                   (return-from load nil))))))))
    (components (name) "Display the components of a remembered run"
     ,#'(lambda (n)
          (format t "~&~{ ~A~}~%"
                  (mapcar #'(lambda (c) (if (equal "" (cddr c)) "-" (cddr c)))
                          (history-components
                           (strategy-history (lookup-remembered n)))))))
    (display-simple (name) "Display a remembered run"
     ,#'(lambda (n)
          (print-strategy-history #'simple-printer (lookup-remembered n) t)))
    (display-tabular (name) "Display a remembered run in tabular form"
     ,#'(lambda (n)
          (print-strategy-history #'tabular-printer (lookup-remembered n) t)))
    (print-eps (name filename)
     "Print a remembered run as EPS"
     ,#'(lambda (n f)
          (with-open-file (out f :direction :output)
            (print-strategy-history #'eps-printer
                                    (lookup-remembered n) out))))
    (display (name) "Display a remembered run as graphics"
     ,#'(lambda (name) (display-run (lookup-remembered name))))
    (decompile (name) "Decompile a remembered strategy"
     ,#'(lambda (name) 
          (format t "~S" (decompile-strategy (lookup-remembered name)))))
    )
  "The list of top-level directives for the GWB.
Entries are of the form (NAME ARGLIST DOC DEFINITION)")

(defun gwb-print-herald ()
  (format t "~&Welcome to the GWB on ~A.~%~
              Type `help' for help.~%"
          (print-universal-time-words nil (get-universal-time))))

(defun repl-1 (&optional (stream *standard-input*))
  (when *gwb-prompt*
    (format t "~&~A" *gwb-prompt*))
  (let ((form (read stream nil)))
    (if form
        (progn (do-form form) t)
      nil)))

(defun do-form (form)
  (cond
    ((symbolp form)
     (let ((def (assoc form *directives*)))
       (when (or (null def) (not (null (cadr def))))
         (error 'gwb-cell-error :kind "directive" :name form))
       (funcall (cadddr def))))
    ((atom form)
     (error 'gwb-syntax-error :form form))
    (t
     (let ((def (assoc (car form) *directives*)))
       (when (or (null def) (null (cadr def)))
         (error 'gwb-syntax-error :form form))
       (when (not (eql (length (cdr form))
                       (length (cadr def))))
         (error 'gwb-syntax-error :form form))
       (apply (cadddr def) (cdr form))))))

(defun gwb ()
  "The top-level of the Games Workbench"
  (let ((*package* (find-package "GWB")))
    (gwb-print-herald)
    (catch 'gwb-quit
      (loop
       (with-simple-restart (gwb "Return to the GWB Top-Level")
         (check-running-processes)
         (if *gwb-handle-errors*
             (handler-case
              (repl-1)
              (gwb-error (c)
                (force-output *standard-output*)
                (format *error-output* "~&ERROR: ~A~%" c)
                (invoke-restart 'gwb)))
           (repl-1))))))
  (kill-running-processes))
